/*******************************************************************************
***  This file is part of Geometry.                                          ***
***                                                                          ***
***  Copyright (C) 2024                                                      ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PROCEDURAL_CARTESIAN_H
#define PROCEDURAL_CARTESIAN_H

#include <stdint.h>


/// The type to use for storing the value of a Cartesian location.
typedef int32_t CARTESIAN_VALUE_TYPE;

/**
 * @brief Representation of a one dimensional location on the Cartesian co-ordinate system.
 */
struct CoordinateAxes
{
	// The discrete location on an axis line, whose absolute value is the distance between it and
	// the origin of that axis. The sign is relative to the origin, i.e. before is negative and
	// after is positive.
	CARTESIAN_VALUE_TYPE value_;
};
/**
 * @brief Representation of a two dimensional location on the Cartesian co-ordinate system.
 */
struct Point
{
	// The horizontal location, also known as the "X" co-ordinate.
	struct CoordinateAxes abscissa_;
	// The vertical location, also known as the "Y" co-ordinate.
	struct CoordinateAxes ordinate_;
};

/**
 * @brief Reads in a one dimensional location from the standard input. The expected input form:
 * value
 *
 * @param location The object to store the read in value.
 *
 * @post The standard input buffer is advanced from reading the value.
 *
 * @return The number of values successfully read or zero if there is a failure. Additionally, EOF
 * may be returned in the event of failure.
 */
int scanfCoordinateAxes(struct CoordinateAxes* location);
/**
 * @brief Reads in a two dimensional location from the standard input. The expected input form:
 * abscissa ordinate
 *
 * @param location The object to store the read in values.
 *
 * @post The standard input buffer is advanced from reading the values.
 *
 * @return The number of values successfully read or zero if there is a failure. Additionally, EOF
 * may be returned in the event of failure.
 */
int scanfPoint(struct Point* location);
/**
 * @brief Writes a one dimensional location to the standard output buffer.
 *
 * @param location The object to write out.
 *
 * @post The standard output buffer is filled with the location value in the form:
 * value
 *
 * @return The number of characters written to the standard output buffer, on error a negative
 * value.
 */
int printfCoordinateAxes(const struct CoordinateAxes location);
/**
 * @brief Writes a two dimensional location to the standard output buffer.
 *
 * @param location The object to write out.
 *
 * @post The standard output buffer is filled with the location values in the form:
 * (abscissa, ordinate).
 *
 * @return The number of characters written to the standard output buffer, on error a negative
 * value.
 */
int printfPoint(const struct Point location);
/**
 * @brief Moves a point horizontally.
 *
 * @param location The object whose position is to be moved.
 * @param amount The amount of units to move horizontally.
 */
void movePointHorizontally(struct Point* location, const CARTESIAN_VALUE_TYPE amount);
/**
 * @brief Moves a point vertically.
 *
 * @param location The object whose position is to be moved.
 * @param amount The amount of units to move vertically.
 */
void movePointVertically(struct Point* location, const CARTESIAN_VALUE_TYPE amount);
/**
 * @brief Moves a point both horizontally and vertically.
 *
 * @param location The object whose position is to be moved.
 * @param horizontalAmount The amount of units to move horizontally.
 * @param verticalAmount The amount of units to move vertically.
 */
void movePoint(
        struct Point* location,
        const CARTESIAN_VALUE_TYPE horizontalAmount,
        const CARTESIAN_VALUE_TYPE verticalAmount
);

#endif
