/*******************************************************************************
***  This file is part of Geometry.                                          ***
***                                                                          ***
***  Copyright (C) 2024                                                      ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include <stdbool.h>
// bool
// false
// true
#include <stdint.h>
// int32_t
#include <stdio.h>
// fprintf
// printf
// puts
// scanf
#include <stdlib.h>
// EXIT_FAILURE
// EXIT_SUCCESS

#include "Shape.h"
// Circle
// Line
// mallocShape
// printfCircle
// printfLine
// putsAllShapes
// scanfCircle
// scanfLine
// ShapeTypes
// toShapeTypes


int populateShapes(enum ShapeTypes types[], void* shapes[], const int32_t maximumShapes);
void freeShapes(void* shapes[], const int32_t numberOfShapes);
int readInShape(const enum ShapeTypes type, void* shape);
bool writeOutShape(const enum ShapeTypes type, const void* shape);
int moveShapes(const enum ShapeTypes types[], void* shapes[], const int numberOfShapes);
bool moveShape(
        const enum ShapeTypes type,
        void* shape,
        const CARTESIAN_VALUE_TYPE horizontalAmount,
        const CARTESIAN_VALUE_TYPE verticalAmount
);

#define HANDLE_ERROR(message)  \
	do  \
	{  \
		perror(message);  \
		freeShapes(shapes, numberOfShapes);  \
		return EXIT_FAILURE;  \
	}  \
	while (0)

int main(int argc, char* argv[])
{
	(void)argc, (void)argv;
	const int32_t shapesSize = 256;
	enum ShapeTypes types[shapesSize];
	void* shapes[shapesSize];
	const int32_t numberOfShapes = populateShapes(types, shapes, shapesSize);

	if (numberOfShapes == -1)
		HANDLE_ERROR("Failed to populate shapes.");

	const int status = moveShapes(types, shapes, numberOfShapes);

	if (status < 0)
		return EXIT_FAILURE;

	freeShapes(shapes, numberOfShapes);
	return EXIT_SUCCESS;
}

int populateShapes(enum ShapeTypes types[], void* shapes[], const int32_t maximumShapes)
{
	enum ShapeTypes selection;
	int32_t nextIndex = 0;

	do
	{
		if (printfAllShapes() < 0)
		{
			perror("Failed to write all shapes.");
			return -1;
		}

		if (puts("Select a shape (Unknown to exit):") < 0)
		{
			perror("Failed to write prompt message.");
			return -1;
		}

		int rawSelection;

		if (scanf("%i", &rawSelection) <= 0)
		{
			perror("Failed to read selection.");
			return -1;
		}

		selection = toShapeTypes(rawSelection);

		if (selection > Unknown)
		{
			types[nextIndex] = selection;
			shapes[nextIndex] = mallocShape(selection);
			const int result = readInShape(types[nextIndex], shapes[nextIndex]);

			if (result <= 0)
			{
				free(shapes[nextIndex]);

				if (result == INVALID_SHAPE_DATA)
				{
					if (fprintf(stderr, "Shape contains invalid data.\n") < 0)
						return -1;
				}
				else if (fprintf(stderr, "Failure while reading in shape data.\n") < 0)
					return -1;

				continue;
			}
			else
				++nextIndex;
		}

		if (nextIndex == maximumShapes)
		{
			if (puts("Maximum number of shapes reached.") < 0)
			{
				perror("Failed to write size error message.");
				return -1;
			}

			break;
		}
	}
	while (selection != Unknown);

	return nextIndex;
}

void freeShapes(void* shapes[], const int32_t numberOfShapes)
{
	for (int32_t index = 0 ; index < numberOfShapes ; ++index)
		free(shapes[index]);
}

int readInShape(const enum ShapeTypes type, void* shape)
{
	switch (type)
	{
	case Unknown:
		return -1;

	case Circle:
		if (puts("Enter in a circle separated by spaces, centre-abscissa centre-ordinate radius:") < 0)
			return -1;

		return scanfCircle((struct Circle*)shape);

	case Line:
		if (puts("Enter in a line separated by spaces, starting-abscissa starting-ordinate ending-abscissa ending-ordinate:") < 0)
			return -1;

		return scanfLine((struct Line*)shape);

	case Rectangle:
		if (puts("Enter in a rectangle separated by spaces, left-upper-corner-abscissa left-upper-corner-ordinate length width:") < 0)
			return -1;

		return scanfRectangle((struct Rectangle*)shape);

	case Square:
		if (puts("Enter in a square separated by spaces, left-upper-corner-abscissa left-upper-corner-ordinate length:") < 0)
			return -1;

		return scanfSquare((struct Square*)shape);

	case Triangle:
		if (puts("Enter in a triangle separated by spaces, first-vertex-abscissa first-vertex-ordinate second-vertex-abscissa second-vertex-ordinate third-vertex-abscissa third-vertex-ordinate:") < 0)
			return -1;

		return scanfTriangle((struct Triangle*)shape);

	default:
		return -1;
	};
}

bool writeOutShape(const enum ShapeTypes type, const void* shape)
{
	switch (type)
	{
	case Unknown:
		return true;

	case Circle:
		if (printf("Circle: ") <= 0)
			return false;

		if (printfCircle(*(const struct Circle*)shape) <= 0)
			return false;

		if (printf("\n") <= 0)
			return false;

		return true;

	case Line:
		if (printf("Line: ") <= 0)
			return false;

		if (printfLine(*(const struct Line*)shape) <= 0)
			return false;

		if (printf("\n") <= 0)
			return false;

		return true;

	case Rectangle:
		if (printf("Rectangle: ") <= 0)
			return false;

		if (printfRectangle(*(const struct Rectangle*)shape) <= 0)
			return false;

		if (printf("\n") <= 0)
			return false;

		return true;

	case Square:
		if (printf("Square: ") <= 0)
			return false;

		if (printfSquare(*(const struct Square*)shape) <= 0)
			return false;

		if (printf("\n") <= 0)
			return false;

		return true;

	case Triangle:
		if (printf("Triangle: ") <= 0)
			return false;

		if (printfTriangle(*(const struct Triangle*)shape) <= 0)
			return false;

		if (printf("\n") <= 0)
			return false;

		return true;

	default:
		return false;
	};
}

int moveShapes(const enum ShapeTypes types[], void* shapes[], const int numberOfShapes)
{
	int selection;

	do
	{
		if (puts("All available shapes:") <= 0)
			HANDLE_ERROR("Failed to display all shapes.");

		for (int32_t index = 0 ; index < numberOfShapes ; ++index)
		{
			if (printf("%i - ", index + 1) <= 0)
				HANDLE_ERROR("Failed to write out shape index.");

			if ( ! writeOutShape(types[index], shapes[index]))
				HANDLE_ERROR("Failed to write out shapes.");
		}

		if (puts("Select shape to move (-1 to exit): ") <= 0)
			HANDLE_ERROR("Failed to write out shape selection request.");

		if (scanf("%i", &selection) <= 0)
		{
			perror("Failed to read move selection.");
			return -1;
		}

		if (selection == -1)
			continue;

		if (selection > numberOfShapes)
		{
			if (puts("Selection exceeds number of available shapes.") <= 0)
				HANDLE_ERROR("Failed to write out shape error message.");

			continue;
		}

		if (printf("Enter horizontal and vertical amount separated by spaces: ") <= 0)
			HANDLE_ERROR("Failed to write out horizontal message.");

		int horizontal, vertical;

		if (scanf("%i %i", &horizontal, &vertical) <= 0)
		{
			perror("Failed to read horizontal and vertical amount.");
			return -1;
		}

		if ( ! moveShape(types[selection - 1], shapes[selection - 1], horizontal, vertical))
			if (puts("Failed to move the shape.") <= 0)
				HANDLE_ERROR("Failed to write out move error message.");
	}
	while (selection != -1);

	return 0;
}

#define MOVE_SHAPE(point)  \
	if (horizontalAmount != 0 && verticalAmount != 0)  \
		movePoint(&point, horizontalAmount, verticalAmount);  \
	else if (horizontalAmount != 0)  \
		movePointHorizontally(&point, horizontalAmount);  \
	else  \
		movePointVertically(&point, verticalAmount)

bool moveShape(
        const enum ShapeTypes type,
        void* shape,
        const CARTESIAN_VALUE_TYPE horizontalAmount,
        const CARTESIAN_VALUE_TYPE verticalAmount
)
{
	switch (type)
	{
	case Unknown:
		return true;

	case Circle:
		MOVE_SHAPE(((struct Circle*)shape)->centre_);
		return true;

	case Line:
		MOVE_SHAPE(((struct Line*)shape)->start_);
		return true;

	case Rectangle:
		MOVE_SHAPE(((struct Rectangle*)shape)->leftEdge_);
		return true;

	case Square:
		MOVE_SHAPE(((struct Square*)shape)->leftEdge_);
		return true;

	case Triangle:
		MOVE_SHAPE(((struct Triangle*)shape)->vertex0_);
		return true;

	default:
		return false;
	};
}

#undef MOVE_SHAPE
